/****************************************************************/
/* gxsmad - provide host information to monitor hosts		*/
/****************************************************************/
/* History:							*/
/* (02-10-2005) v0.8.3 Added code for v2 protocol stuff. Todd	*/
/* (02-27-2002) v0.8.2 Added code to allow IP address or	*/
/*		hostname in 'HOST' entry in config file. Todd	*/
/* (12-16-2001) v0.8.1 Added protocol negotiation code. Todd	*/
/* (11-27-2001) v0.8 Add OpenSSL support for encryption. Todd	*/
/* (03-06-2001) v0.7 Compile as a service under W32. Todd	*/
/* (09-07-2000) v0.5 Port to win32 complete. Todd Shadburn	*/
/* (08-18-2000) v0.4.1 Begin port to win32 platform. Added the	*/
/*		clientparms stuff to easy porting for different	*/
/*		threading structures. Todd			*/
/* (09-13-1999) v0.4 Added the commandline stuff.Todd		*/
/* (09-02-1999) v0.3 Fix the bug where the cli_socket was left	*/
/*		in a CLOSE_WAIT state after the conversation	*/
/*		was over. Added 'delete_lists()'. Todd Shadburn	*/
/* (08-31-1999) v0.2 Changes to config file to specify both	*/
/*		allowed hosts and allowed commands.Todd		*/
/* (08-24-1999) Orginal creation date. Todd Shadburn		*/
/****************************************************************/


#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
/*
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>
*/
#else
#include <wtypes.h>
#include <process.h>
#include <winsock2.h>
#include <winsvc.h>
#include <io.h>
#endif

#include <signal.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>

#include <cutil/cutil.h>
#include <comm/comm.h>

#include <gxsma/gxsma_portable.h>
#include <gxsma/gxsma.h>
#include <gxsma/nmodule_p.h>
#include <gxsma/conffile.h>
#include <gxsma/event.h>
#include <gxsma/event_p.h>


/****************************************************************/
/* Preprocessor definitions					*/
/****************************************************************/

#define TEMPFILE_NAME_SIZE	64
#define ARGUMENT_IDENTIFIER	"$ARGS$"

#define GXSMAD_LOG_ENTRY_SIZE	128

#define GXSMAD_FLAG_NODAEMON	0x00000001

#define GXSMAD_HEARTBEAT_INTERVAL	60

#define GXSMAD_FILEPATH_SIZE	128
#define GXSMAD_CONFIG_FILE	"gxsmad.conf"
#define GXSMAD_LOG_FILE	"gxsmad.log"

#ifdef WIN32
#define	MSG_STARTED		"Gxsmad has started."
#define	MSG_ENDED		"Gxsmad has ended."
#define W32_SERVICE		1
#endif

#define IO_BUFFER_SIZE		4096

#define ADDED_EVENT_CONCAT_20050827
#ifdef ADDED_EVENT_CONCAT_20050827
#define GXSMA_EVENTCHAIN_TIMEOUT_SECONDS	120
#endif

#ifdef W32_SERVICE
#ifdef __GNUC__
#define RTMAIN			int
#define RETURNMAIN		return 0
#else
#define RTMAIN			VOID
#define RETURNMAIN		return
#endif
#endif


#ifndef WIN32
#define socket_t int
#define close_socket(s)	close(s)
#define open_file(f, a, m) open(f, a, m)
#define close_file(f) close(f)
#define INVALID_SOCKET -1
#define THREAD_EXIT(x)	exit(x)
#else
#define socket_t SOCKET
#define close_socket(s)	closesocket(s)
#define open_file(f, a, m) _open(f, a)
#define close_file(f) _close(f)
#define unlink(f) DeleteFile(f)
#define read(f, b, l) _read(f, b, l)
#define THREAD_EXIT(x)	ExitThread(x)
#define stat(x, y)	_stat(x, y)
#endif

/****************************************************************/
/* Structure Prototypes						*/
/****************************************************************/
typedef struct
	{
	socket_t		clientsocket;
	struct sockaddr_in	*clientsockaddr;
	int			sockaddrsize;
	/* added for protocol v2 */
	CommChannel		*cc;
	} clientparms_t;


typedef struct
	{
	u_int32_t	serverflags;
	SList		*HostList;
	SList		*CommandList;
	SList		*cfgmodulelist;
	SList		*ModuleList;
	unsigned short	serverport;
	char		config_file[GXSMAD_FILEPATH_SIZE];
	char		log_file[GXSMAD_FILEPATH_SIZE];
	FILE		*lfd;
	pthread_mutex_t	log_mutex;
	char		*pubkeyfile;
	char		*privkeyfile;
	char		*pemphrase;
	ModuleVector	modvectors;
	char		*eventhost;
	unsigned short	eventport;
	CommModule	*cm;
	CommChannel	*ccs;
	Attribute	*commattribs;
#ifdef ADDED_EVENT_CONCAT_20050827
	CommChannel	*eventcc;
	u_int32_t	lasteventtime;
	u_int32_t	eventtimeout;
#endif
	u_int32_t	serverpid;
	CommGroup	*cg;
	u_int32_t	conncount;
	EventClient	*ec;
	} serverparms_t;
	

/****************************************************************/
/* Global variables						*/
/****************************************************************/
/*
char unknown_message[] = "I don't talk to strangers.\n";
char unknown_request[] = "I don't know that command.\n";
*/

serverparms_t		Server;

#ifdef W32_SERVICE
SERVICE_STATUS		ServiceStatus;
SERVICE_STATUS_HANDLE	ServiceStatusHandle;
#endif

#define BUFFER_SIZE		256

/****************************************************************/
/* Function Prototypes						*/
/****************************************************************/
void signal_handler(int sig);
void handle_client(clientparms_t *clientparms);
int access_file_load(char *filename);
void delete_lists(void);
int host_has_access(long hostname);
/*char * get_command_from_request(char * cli_request); */
int pem_passwd_cb(char *buf, int size, int rwflag, void *password);
int send_response_v2 (CommChannel *cc,
			u_int32_t status,
			u_int32_t flags,
			u_int32_t responserc,
			char *command,
			char *data,
			u_int32_t datalen);
int gxsmad_log_entry (char *format, ...);
/* int gxsmad_method_marshall (clientparms_t * cp, char *cmd, GxsmaHeader *nh); */
int gxsmad_vector_log_entry (ModuleInfo *mi, char *format, va_list args);
int gxsmad_vector_emit_event (ModuleInfo *mi, EventData *ed);
/* int gxsmad_emit_event (EventData *ed); */
int gxsmad_module_load (clientparms_t *cp, GxsmaHeader *nh);
int gxsmad_module_unload (clientparms_t *cp, GxsmaHeader *nh);
int gxsmad_vector_eventchain_open (ModuleInfo *mi, EventChain **ec);
int gxsmad_vector_eventchain_close (EventChain **ec);
int gxsmad_vector_eventchain_emit_event (EventChain *ec, EventData *ed);
#ifdef ADDED_EVENT_CONCAT_20050827
void *gxsmad_housekeeper_thread (void * arg);
#endif
/* int gxsmad_restart (clientparms_t *cp, GxsmaHeader *nh); */
int gxsmad_restart (void);
int handle_connection_io (CommChannel *cc, int iotype, void *ccarg);


#ifndef W32_SERVICE
void usage(void);
#endif

#ifdef W32_SERVICE
VOID GxsmadServiceStart(DWORD argc, LPTSTR *argv);
VOID GxsmadServiceCtrlHandler(DWORD opcode);

RTMAIN main(void)
	{
	char str[256];
	SERVICE_TABLE_ENTRY DispatchTable[] =
		{
			{"Gxsmad",
			  (LPSERVICE_MAIN_FUNCTION)GxsmadServiceStart},
			{NULL, NULL}
		};
	if(!StartServiceCtrlDispatcher(DispatchTable))
		{
		sprintf(str,
			"Gxsmad: StartServiceCtrlDispatcher error = %d\n",
			GetLastError());
		OutputDebugStringA(str);
		}
	RETURNMAIN;
	}
#endif


/****************************************************************/
/* main - Start of execution					*/
/****************************************************************/
#ifdef WIN32
#ifdef W32_SERVICE
void GxsmadServiceStart(DWORD argc, LPTSTR *argv)
#else
int main(DWORD argc, char * argv[])
#endif
#else
int main(int argc, char * argv[])
#endif
{
	int		rc;
#ifndef WIN32
	int		option;
	pid_t		pid;
#else
	DWORD		option;
	WSADATA		wsaData;
	char		str[256];
#endif

	char		*modpath = NULL;
	SList		*litem = NULL;
	GxsmaModule	*nm = NULL;
	EventData	*event = NULL;

	memset (&Server, 0, sizeof (Server));
	Server.serverflags = 0;
	Server.serverport = 0;	// Disable the local listener by default
				//  (use the 'PORT' config file option to enable it)
	Server.pubkeyfile = NULL;
	Server.privkeyfile = NULL;
	Server.modvectors.log_entry = gxsmad_vector_log_entry;
	Server.modvectors.emit_event = gxsmad_vector_emit_event;
	Server.modvectors.eventchain_open = gxsmad_vector_eventchain_open;
	Server.modvectors.eventchain_close = gxsmad_vector_eventchain_close;
	Server.modvectors.eventchain_emit_event =
				gxsmad_vector_eventchain_emit_event;
	Server.eventport = NIEVENT_DEFAULT_PORT;
#ifdef ADDED_EVENT_CONCAT_20050827
	Server.eventtimeout = GXSMA_EVENTCHAIN_TIMEOUT_SECONDS;
#endif

#ifdef W32_SERVICE
	ServiceStatus.dwServiceType		= SERVICE_WIN32;
	ServiceStatus.dwCurrentState		= SERVICE_START_PENDING;
	ServiceStatus.dwControlsAccepted	= SERVICE_ACCEPT_STOP;
	ServiceStatus.dwWin32ExitCode		= 0;
	ServiceStatus.dwServiceSpecificExitCode = 0;
	ServiceStatus.dwCheckPoint		= 0;
	ServiceStatus.dwWaitHint		= 0;
	ServiceStatusHandle = RegisterServiceCtrlHandler("Gxsmad",
			(LPHANDLER_FUNCTION)GxsmadServiceCtrlHandler);
	if(ServiceStatusHandle == 0) {
		sprintf(str,
			"Gxsmad: RegisterServiceCtrlHandler error = %d\n",
			GetLastError());
		OutputDebugStringA(str);
	}
#endif

	gxsma_get_component_filepath (Server.config_file, GXSMAD_FILEPATH_SIZE, "etc", GXSMAD_CONFIG_FILE);
	gxsma_get_component_filepath (Server.log_file, GXSMAD_FILEPATH_SIZE, "log", GXSMAD_LOG_FILE);

#ifndef WIN32
	while (( option = getopt(argc, argv, "Df:hl:V")) != EOF) {
		switch(option) {
			case 'D':
				Server.serverflags |= GXSMAD_FLAG_NODAEMON;
				fprintf (stderr, "Not daemonizing.\n");
				break;
			case 'f':
				strcpy(Server.config_file, optarg);
				break;
			case 'l':
				strcpy(Server.log_file, optarg);
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'V':
				printf("\ngxsmad - version %s\n\n",
					VERSION);
				exit(0);
				break;
			default:
				usage();
				exit(-1);
				break;
		}
	}
#else
	option = 1;
	while (option < argc) {
		if(argv[option][0] == '-') {
			switch(argv[option][1]) {
				case 'f':
					if((option+1) < argc) {
						option++;
						strcpy(Server.config_file, argv[option]);
					}
					break;
				case 'l':
					if((option+1) < argc) {
						option++;
						strcpy(Server.log_file, argv[option]);
					}
					break;
#ifndef W32_SERVICE
				case 'h':
					usage();
					exit(0);
					break;
				case 'V':
					printf("\ngxsmad - version %s\n\n",
						VERSION);
					exit(0);
					break;
				default:
					usage();
					exit(-1);
					break;
#else
				default:
					exit(-1);
					break;
#endif
			}
		} else {
#ifndef W32_SERVICE
			usage();
#endif
			exit(-1);
		}
		option++;
	}
#endif

#ifndef WIN32
	if (!(Server.serverflags & GXSMAD_FLAG_NODAEMON)) {
		if ( (pid=fork()) < 0)
			exit(-1);
		else if(pid != 0)
			exit(0);
		chdir ("/");
		setsid ();
		umask (0);
		close (2);
		close (1);
		close (0);
		Server.serverpid = getpid();
	}
#endif

	rc = pthread_mutex_init (&Server.log_mutex, NULL);
	Server.lfd=fopen(Server.log_file, "a");
	if(!Server.lfd) {
		exit(-4);
	}
	gxsmad_log_entry ("gxsmad starting...\n");

#ifdef WIN32
	WSAStartup(MAKEWORD(2,0), &wsaData);
#endif
	rc = access_file_load(Server.config_file);
	if(rc) {
		gxsmad_log_entry ("Error loading config file. rc=%d\n", rc);
		exit(-1);
	}

	rc = comm_plugin_load (&Server.cm,
			"comm_openssl",
			Server.commattribs);
	if(rc != 0) {
		gxsmad_log_entry ("comm_plugin_load failed. rc=%d\n", rc);
		exit(-1);
	}

	litem = Server.cfgmodulelist;
	while (litem != NULL) {
		modpath = (char *)slist_get_data (litem);
		gxsmad_log_entry ("Loading module '%s'\n", modpath);
		rc = gxsma_module_load (&nm,
					modpath,
					&Server.modvectors);
		if (rc != GXSMA_RC_SUCCESS) {
			gxsmad_log_entry (
					"Failed to load module. rc=%d\n", rc);
		} else {
			slist_item_add(&Server.ModuleList, (void *)nm);
		}
		litem = slist_get_next (litem);
	}

	rc = comm_group_new(&Server.cg, handle_connection_io);
	if(rc != 0) {
		gxsmad_log_entry ("comm_group_new failed. rc=%d\n", rc);
		rc = comm_plugin_unload (&Server.cm);
		exit(-1);
	}

	if (Server.serverport != 0) {
		gxsmad_log_entry ("Starting local listener on port %d\n", Server.serverport);
		rc = comm_bind (Server.cm, &Server.ccs, 1, Server.serverport);
		if(rc != 0) {
			gxsmad_log_entry ("comm_bind failed. rc=%d\n", rc);
			rc = comm_group_free (&Server.cg);
			rc = comm_plugin_unload (&Server.cm);
			exit(-1);
		}
		
		rc = comm_group_add (Server.cg, Server.ccs, NULL, COMM_IO_READ);
		if(rc != 0) {
			gxsmad_log_entry ("comm_group_add failed. rc=%d\n", rc);
			rc = comm_close (&Server.ccs, 0);
			rc = comm_group_free (&Server.cg);
			rc = comm_plugin_unload (&Server.cm);
			exit(-1);
		}
	} else {
		gxsmad_log_entry ("No local listener will be started.\n");
	}

#ifndef WIN32
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP, signal_handler);
	signal(SIGPIPE, signal_handler);
	signal(SIGCHLD, SIG_IGN);
#endif
#ifdef W32_SERVICE
	ServiceStatus.dwCurrentState	= SERVICE_RUNNING;
	ServiceStatus.dwCheckPoint	= 0;
	ServiceStatus.dwWaitHint	= 0;
	if(!SetServiceStatus(ServiceStatusHandle, &ServiceStatus)) {
		sprintf(str, "Gxsmad: SetServiceStatus error = %d\n",
				GetLastError());
		OutputDebugStringA(str);
	}
#endif


	rc = eventclient_init (&Server.ec, 0);
	if(rc != 0) {
		gxsmad_log_entry ("eventclient_init failed. rc=%d\n", rc);
		rc = comm_close (&Server.ccs, 0);
		rc = comm_group_free (&Server.cg);
		rc = comm_plugin_unload (&Server.cm);
		exit(-1);
	}
	rc = eventclient_configure_heartbeat (Server.ec,
					"GXSMAD",
					GXSMAD_HEARTBEAT_INTERVAL);
	rc = eventclient_set_option (Server.ec,
					EC_OPTION_LOGGING_FUNCTION,
					gxsmad_log_entry);
	rc = eventdata_new (&event);
	rc = eventdata_set (event,
			0,
			0,
			"SYSEVENT",
			"Gxsmad agent started",
			strlen ("Gxsmad agent started"));
/*	rc = gxsmad_emit_event (event); */
	rc = eventclient_emit_event (Server.ec, event);
	rc = eventdata_free (&event);


	rc = comm_group_loop (Server.cg);

	rc = eventclient_uninit (&Server.ec);
	rc = comm_close (&Server.ccs, 0);
	rc = comm_plugin_unload (&Server.cm);
	/*rc = openssl_comm_cleanup (&Server.cmd);*/
	attribute_freeall (&Server.commattribs);
#ifdef WIN32
	WSACleanup();
#endif
	exit(0);
#ifdef W32_SERVICE
	return;
#else
	return(0);
#endif
}


#ifndef WIN32
void signal_handler(int sig)
{
	int rc;
	EventData *event = NULL;
	switch(sig) {
		case SIGPIPE:
			/*gxsmad_log_entry ("SIGPIPE: Attempted write to an unconnected socket.\n"); */
			break;
		case SIGCHLD:
			wait(&rc);
			break;
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:

			/*
			 * This is too much in the signal handler and is
			 * probably causing the hangs (__lll_lock_wait_private).
			 * It needs to be moved out! (unsafe, see CWE-364)...
			*/
			/* ... so cheat, for now. */
			exit (0);

			rc = comm_close (&Server.ccs, 0);
			rc = eventdata_new (&event);
			rc = eventdata_set (event,
					0,
					2,
					"SYSEVENT",
					"Gxsmad agent ended",
					strlen ("Gxsmad agent ended"));
			rc = eventclient_emit_event (Server.ec, event);
			/* rc = gxsmad_emit_event (event); */
			rc = eventdata_free (&event);

			rc = comm_plugin_unload (&Server.cm);
			/*openssl_comm_cleanup (&Server.cmd);*/
			delete_lists();
			/*
			free(Server.pubkeyfile);
			free(Server.privkeyfile);
			*/
			gxsmad_log_entry ("Server shutdown.\n");
			fflush(Server.lfd);
			fclose(Server.lfd);
			exit(0);
			break;
		case SIGHUP:
			/*
			 * Not complete and unsafe (CWE-364)
			gxsmad_log_entry ("SIGHUP: Reloading Server.config file.\n");
			delete_lists();
			access_file_load(Server.config_file);
			*/
			break;
	}
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP, signal_handler);
	signal(SIGPIPE, signal_handler);
	return;
}
#endif


void handle_client(clientparms_t *clientparms)
{
	int rc;
	int cli_size;
	long cli_ipaddress;
	char cli_addr[20];
	socket_t cli_socket;
	struct sockaddr_in * cli_sockaddr = NULL;
	if (clientparms == NULL || clientparms->cc == NULL) {
		free (clientparms->cc->cmodule);
		free (clientparms->cc);
		free (clientparms->clientsockaddr);
		free (clientparms);
		THREAD_EXIT(0);
		return ;
	}

	gxsmad_log_entry ("Entered handle_client\n");
	cli_socket = clientparms->clientsocket;
	cli_size = clientparms->sockaddrsize;
	cli_sockaddr = clientparms->clientsockaddr;
	cli_ipaddress = htonl(cli_sockaddr->sin_addr.s_addr);
	sprintf(cli_addr, "%d.%d.%d.%d",
		(unsigned int)cli_ipaddress >> 24,
		(unsigned int)(cli_ipaddress << 8) >> 24,
		(unsigned int)(cli_ipaddress << 16) >> 24,
		(unsigned int)(cli_ipaddress << 24) >> 24);
	if(!host_has_access(cli_ipaddress)) {
		gxsmad_log_entry ("Attempted Connection from: %s\n", cli_addr);
		rc = comm_close (&clientparms->cc, 0);
		free (clientparms->clientsockaddr);
		free (clientparms);
		THREAD_EXIT(0);
		return;
	}
	gxsmad_log_entry ("Connection accepted from: %s\n", cli_addr);

	// Do work here

	gxsmad_log_entry ("Disconnecting.\n");
	rc = comm_close (&clientparms->cc, 0);
	free (clientparms->clientsockaddr);
	free (clientparms);
	THREAD_EXIT(0);
	return;
}


int access_file_load(char *filename)
{
	int rc;
	int count;
	int len;
	char line[BUFFER_SIZE];
	long host;
	ConfigFile *cf = NULL;
	struct hostent *hentry = NULL;
	char *command = NULL;
	char *directive = NULL;
	char *target = NULL;
	len=120;
	
	rc = configfile_open (&cf, filename);
	if(rc != GXSMA_RC_SUCCESS)
		return(GXSMA_RC_INVALIDARG);
	while(1) {
		rc = configfile_read_line (cf, line, BUFFER_SIZE);
		if (rc != GXSMA_RC_SUCCESS)
			break;
		if(line[0] == '#' || strlen (line) < 2)
			continue;
		directive = strtok (line, " \t\r\n");
		if (directive == NULL)
			continue;
		if(!strcmp(directive, "HOST")) {
			target = strtok (NULL, "\t\r\n");
			if (target == NULL)
				continue;
			hentry=gethostbyname(target);
			if(!hentry) {
				gxsmad_log_entry ("HOST: %s: DNS lookup error.\n",
					target);
				continue;
			}
			count=0;
			while(hentry->h_addr_list[count] != NULL) {
				gxsmad_log_entry ("HOST: %s -> %d.%d.%d.%d added.\n",
					hentry->h_name,
					(unsigned char)hentry->h_addr_list[count][0],
					(unsigned char)hentry->h_addr_list[count][1],
					(unsigned char)hentry->h_addr_list[count][2],
					(unsigned char)hentry->h_addr_list[count][3]);
				host=*((long *)hentry->h_addr_list[count]);
				slist_item_add (&Server.HostList, (void *)host);
				count++;
			}
		} else if(!strcmp(directive, "COMMAND")) {
			target = strtok (NULL, "\t\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			slist_item_add (&Server.CommandList, (void *)command);
		} else if(!strcmp(directive, "PUBLICKEY")) {
			target = strtok (NULL, "\t\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			Server.pubkeyfile = command;
			Server.commattribs = attribute_add (Server.commattribs,
							"SSLCertificateFile",
							command);
		} else if(!strcmp(directive, "PRIVATEKEY")) {
			target = strtok (NULL, "\t\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			Server.privkeyfile = command;
			Server.commattribs = attribute_add (Server.commattribs,
							"SSLPrivateKeyFile",
							command);
		} else if(!strcmp(directive, "PEMPHRASE")) {
			target = strtok (NULL, "\t\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			Server.pemphrase = command;
		} else if(!strcmp(directive, "PORT")) {
			target = strtok (NULL, "\t\r\n");
			if (target == NULL)
				continue;
			Server.serverport=atoi(target);
		} else if(!strcmp(directive, "MODULE")) {
			target = strdup (strtok (NULL, "\t\r\n"));
			if (target == NULL)
				continue;
			slist_item_add(&Server.cfgmodulelist, (void *)target);
/*
			gxsmad_log_entry ("Loading module '%s'\n",
						target);
			rc = gxsma_module_load (&nm,
						target,
						&Server.modvectors);
			if (rc != GXSMA_RC_SUCCESS) {
				gxsmad_log_entry (
					"Failed to load module. rc=%d\n",
					rc);
			} else {
				slist_item_add(&Server.ModuleList, (void *)nm);
			}
*/
		} else if(!strcmp(directive, "EVENTHOST")) {
			target = strdup (strtok (NULL, "\t\r\n"));
			if (target == NULL)
				continue;
			Server.eventhost = target;
		} else if(!strcmp(directive, "EVENTPORT")) {
			target = strtok (NULL, "\t\r\n");
			if (target == NULL)
				continue;
			Server.eventport = atoi (target);
		} else {
			fprintf(stderr, "Invalid directive: '%s'\n", directive);
			configfile_close (&cf);
			return(-2);
		}
	}
	configfile_close (&cf);
	return(0);
}


int host_has_access(long hostname)
{
	long host;
	long nohost;
	SList *litem = NULL;
	litem = Server.HostList;
	while (litem != NULL) {
		host=(long)slist_get_data (litem);
		nohost=htonl(hostname);
		if(nohost == host)
			return(1);
		litem = slist_get_next (litem);
	}
	return(0);
}

/*
char * get_command_from_request(char * cli_request)
{
	char *command = NULL;
	char *item = NULL;
	SList *litem = NULL;
	if (cli_request == NULL)
		return (NULL);
	litem = Server.CommandList;
	while (litem != NULL) {
		item = strdup(slist_get_data (litem));
		command = strtok (item, " \t");
		if (!strcmp (command, cli_request)) {
			command = strdup (strtok (NULL, "\r\n"));
			free (item);
			return command;
		}
		free(item);
		litem = slist_get_next (litem);
	}
	return(NULL);
}
*/

void delete_lists(void)
{
	int rc;
	void *p = NULL;
	SList *litem = NULL;
	GxsmaModule *nm = NULL;
	gxsmad_log_entry ("Freeing module, host & command lists.\n");
	/* HostList is now a list of longs not char*... */
	/* dont try to free the items, just the list */
	slist_free (&Server.HostList);
	litem = Server.CommandList;
	while (litem != NULL) {
		p = slist_get_data (litem);
		free (p);
		litem = slist_get_next (litem);
	}
	slist_free (&Server.CommandList);
	litem = Server.ModuleList;
	while (litem != NULL) {
		nm = (GxsmaModule *)slist_get_data (litem);
		rc = gxsma_module_unload (&nm);
		litem = slist_get_next (litem);
	}
	slist_free (&Server.ModuleList);
	return;
}


int pem_passwd_cb(char *buf, int size, int rwflag, void *password)
{
	int len;
	len = strlen(Server.pemphrase);
	if(len > size)
		len = size - 1;
	strncpy(buf, Server.pemphrase, len);
	buf[len] = 0x00;
	return(len);
}


#ifndef W32_SERVICE
void usage(void)
{
	printf("\ngxsmad - Information Server.n");
	printf(" Options:\n");
#ifndef WIN32
	printf("   -D                   Gxsmad will not daemonize itself.\n");
#endif
	printf("   -f configfile	Specifies the path to the ");
	printf("configuration file to use.\n");
	printf("   -h			This help screen.\n");
	printf("   -l logfile		Specifies the path to the ");
	printf("log file to use.\n");
	printf("   -V			Display version information ");
	printf("and exit.\n\n");
	return;
}
#endif

#ifdef W32_SERVICE
void GxsmadServiceCtrlHandler(DWORD opcode)
{
	CHAR str[256];
	switch(opcode) {
		case SERVICE_CONTROL_STOP:
			ServiceStatus.dwWin32ExitCode	= 0;
			ServiceStatus.dwCurrentState	= SERVICE_STOPPED;
			ServiceStatus.dwCheckPoint	= 0;
			ServiceStatus.dwWaitHint	= 0;
			break;
	}
	if(!SetServiceStatus(ServiceStatusHandle, &ServiceStatus)) {
		sprintf(str, "Gxsmad: SetServiceStatus error = %lu\n",
				GetLastError());
		OutputDebugStringA(str);
	}
	return;
}
#endif


int send_response_v2 (CommChannel *cc,
			u_int32_t status,
			u_int32_t flags,
			u_int32_t responserc,
			char *command,
			char *data,
			u_int32_t datalen)
{
	int rc;
	int bytes;
	int count;
	GxsmaHeader nh;
	char *pdata = NULL;
	if (cc == NULL)
		return GXSMA_RC_INVALIDARG;
	rc = gxsma_header_clear (&nh);
	rc = gxsma_header_set_status (&nh, status);
	rc = gxsma_header_set_rc (&nh, responserc);
	if (command != NULL)
		rc = gxsma_header_set_command (&nh, command);
	if (data != NULL && datalen > 0)
		rc = gxsma_header_set_datalen (&nh, datalen);
	rc = gxsma_header_to_networkorder (&nh);
	bytes = sizeof (GxsmaHeader);
	rc = comm_write(cc, (char*)&nh, &bytes);
	if (data != NULL && datalen > 0) {
		bytes = datalen;
		pdata = data;
		while (bytes > 0) {
			count = (bytes > 4096 ? 4096 : bytes);
			rc = comm_write(cc,
					pdata,
					&count);
			bytes -= count;
			pdata += count;
		}
	}
	return GXSMA_RC_SUCCESS;
}


int gxsmad_log_entry (char *format, ...)
{
	va_list args;
	int rc;
	time_t t;
	struct tm ltr;
	struct tm *lt;
	char *str = NULL;
	if (format == NULL || Server.lfd == NULL)
		return GXSMA_RC_INVALIDARG;
	str = calloc (1, GXSMAD_LOG_ENTRY_SIZE);
	if (str == NULL)
		return GXSMA_RC_MEMORYERROR;
	va_start (args, format);
	t = time (NULL);
	lt = localtime_r (&t, &ltr);
	fprintf (Server.lfd,
		"%04d-%02d-%02d %02d:%02d:%02d ",
		ltr.tm_year+1900,
		ltr.tm_mon+1,
		ltr.tm_mday,
		ltr.tm_hour,
		ltr.tm_min,
		ltr.tm_sec);
	rc = vsnprintf (str,
			GXSMAD_LOG_ENTRY_SIZE,
			format,
			args);
	pthread_mutex_lock (&Server.log_mutex);
	fprintf (Server.lfd, "%s", str);
	pthread_mutex_unlock (&Server.log_mutex);
	fflush (Server.lfd);
	free (str);
	va_end (args);
	return GXSMA_RC_SUCCESS;
}


int gxsmad_vector_log_entry (ModuleInfo *mi, char *format, va_list args)
{
	int rc;
	time_t t;
	struct tm ltr;
	struct tm *lt;
	char *str = NULL;
	if (mi == NULL || format == NULL || Server.lfd == NULL)
		return GXSMA_RC_INVALIDARG;
	str = calloc (1, GXSMAD_LOG_ENTRY_SIZE);
	if (str == NULL)
		return GXSMA_RC_MEMORYERROR;
	t = time (NULL);
	lt = localtime_r (&t, &ltr);
	fprintf (Server.lfd,
		"%04d-%02d-%02d %02d:%02d:%02d (%s) ",
		ltr.tm_year+1900,
		ltr.tm_mon+1,
		ltr.tm_mday,
		ltr.tm_hour,
		ltr.tm_min,
		ltr.tm_sec,
		mi->name);
	rc = vsnprintf (str,
			GXSMAD_LOG_ENTRY_SIZE,
			format,
			args);
	pthread_mutex_lock (&Server.log_mutex);
	fprintf (Server.lfd, "%s", str);
	pthread_mutex_unlock (&Server.log_mutex);
	fflush (Server.lfd);
	free (str);
	return GXSMA_RC_SUCCESS;
}


int gxsmad_module_load (clientparms_t *cp, GxsmaHeader *nh)

{
	int		rc;
	int		lrc;
	int		bytes;
	char		filepath[GXSMA_HEADER_ARGDATA_SIZE];
	GxsmaModule	*nm = NULL;
	if (cp == NULL || nh == NULL)
		return GXSMA_RC_INVALIDARG;
	memset (filepath, 0, GXSMA_HEADER_ARGDATA_SIZE);
	bytes = GXSMA_HEADER_ARGDATA_SIZE;
	rc = gxsma_header_get_argdata (nh, filepath, &bytes);
	gxsmad_log_entry ("module.load: path=%s\n", filepath);
	
	lrc = gxsma_module_load (&nm,
				filepath,
				&Server.modvectors);
	if (lrc != GXSMA_RC_SUCCESS) {
		gxsmad_log_entry (
			"Failed to load module. rc=%d\n",
			lrc);
	} else {
		slist_item_add(&Server.ModuleList, (void *)nm);
	}
	rc = send_response_v2 (cp->cc, 
			GXSMA_STATUS_RESPONSE,
			0,
			lrc,
			"module.load",
			NULL,
			0);
	return GXSMA_RC_SUCCESS;
}


int gxsmad_module_unload (clientparms_t *cp, GxsmaHeader *nh)
{
	int		rc;
	int		lrc;
	int		bytes;
	char		name[GXSMA_HEADER_ARGDATA_SIZE];
	GxsmaModule	*nm = NULL;
	SList		*litem = NULL;
	if (cp == NULL || nh == NULL)
		return GXSMA_RC_INVALIDARG;
	memset (name, 0, GXSMA_HEADER_ARGDATA_SIZE);
	bytes = GXSMA_HEADER_ARGDATA_SIZE;
	rc = gxsma_header_get_argdata (nh, name, &bytes);
	gxsmad_log_entry ("module.unload: name=%s\n", name);
	
	litem = Server.ModuleList;
	while (litem != NULL) {
		nm = slist_get_data (litem);
		if (nm != NULL && nm->moduleinfo.name != NULL &&
			!strcmp (nm->moduleinfo.name, name)) {
			break;
		}
		litem = slist_get_next (litem);
	}
	if (litem == NULL) {
		rc = send_response_v2 (cp->cc, 
				GXSMA_STATUS_RESPONSE |
					GXSMA_STATUS_ARGINVALID,
				0,
				GXSMA_RC_INVALIDARG,
				"module.unload",
				NULL,
				0);
		return GXSMA_RC_INVALIDARG;
	}
	lrc = gxsma_module_unload (&nm);
	if (lrc != GXSMA_RC_SUCCESS) {
		gxsmad_log_entry (
			"Failed to unload module. rc=%d\n",
			lrc);
	} else {
		slist_item_remove(&Server.ModuleList, (void *)nm);
	}
	rc = send_response_v2 (cp->cc, 
			GXSMA_STATUS_RESPONSE,
			0,
			lrc,
			"module.unload",
			NULL,
			0);
	return GXSMA_RC_SUCCESS;
}


/* New for eventclient */

int gxsmad_vector_eventchain_open (ModuleInfo *mi, EventChain **ec)
{
	EventChain	*c = NULL;
	if (mi == NULL || ec == NULL)
		return GXSMA_RC_INVALIDARG;
	/* START just fake it for now, need to alter lib/gxsma_p.c */
	c = calloc (1, sizeof (EventChain));
	if (c == NULL)
		return GXSMA_RC_MEMORYERROR;
	c->moduleinfo = mi;
	*ec = c;
	/* END just fake it for now, need to alter lib/gxsma_p.c */
	return GXSMA_RC_SUCCESS;
}


int gxsmad_vector_eventchain_close (EventChain **ec)
{
	if (ec == NULL || *ec == NULL)
		return GXSMA_RC_INVALIDARG;
	free (*ec);
	*ec = NULL;
	return GXSMA_RC_SUCCESS;
}


int gxsmad_vector_eventchain_emit_event (EventChain *ec, EventData *ed)
{
	int		rc;
	if (ec == NULL || ed == NULL)
		return GXSMA_RC_INVALIDARG;
	rc = eventclient_emit_event (Server.ec, ed);
	return rc;
}


int gxsmad_vector_emit_event (ModuleInfo *mi, EventData *ed)
{
	int		rc;
	if (mi == NULL || ed == NULL)
		return GXSMA_RC_INVALIDARG;
	rc = eventclient_emit_event (Server.ec, ed);
	return rc;
}



#ifdef ADDED_EVENT_CONCAT_20050827
void *gxsmad_housekeeper_thread (void * arg)
{
	int		seconds;
	while (1) {
		sleep (Server.eventtimeout);
		seconds = time (NULL) - Server.lasteventtime;
		if (seconds > Server.eventtimeout) {
			/* Close the event CommChannel */
		}
	}
#ifndef WIN32
	pthread_exit (NULL);
#else
	THREAD_EXIT(0);
#endif
	return (NULL);
}
#endif


#ifndef WIN32
int gxsmad_restart (void)
{
/*
	int	rc;
	int	trc;
	pid_t	pid;
	gxsmad_log_entry (
		"Restarting gxsmad...\n");
	if ( (pid=fork()) < 0)
		return errno;
	else if (pid != 0) {
		pid = getpid();
		rc = kill (Server.serverpid, SIGTERM);
		return GXSMA_RC_SUCCESS;
	}
	gxsmad_log_entry (
		"(restart pid=%d) waiting for gxsmad(pid=%d) to exit...\n",
		getpid(),
		Server.serverpid);
	rc = waitpid (Server.serverpid, &trc, 0);
	chdir ("/");
	setsid ();
	close (2);
	close (1);
	close (0);
	delete_lists();
	gxsmad_log_entry (
		"(restart) restarting server...\n", getpid());
	fflush(Server.lfd);
	fclose(Server.lfd);
	sleep (2);
	rc = system (SBINDIR DIR_DELIMITER "gxsmad");
	exit (0);
*/
	return 0;
}
#else
int gxsmad_restart (void)
{
	/* Not supported on win32 platform */
	return 0;
}
#endif



int handle_connection_io (CommChannel *cc, int iotype, void *ccarg)
{
        int                     rc;
        CommChannel             *ccc = NULL;
	clientparms_t		*clientparms = NULL;
        if (cc == NULL)
		return EINVAL;
        // If it's the server socket, accept the new connection
        if (cc == Server.ccs) {
                rc = comm_accept (cc, &ccc, 0);
                if (rc != 0) {
                        gxsmad_log_entry (
                                "comm_accept(): failed rc=%d\n", rc);
                        return rc;
                }
		// allocate client structures
		clientparms = calloc (1, sizeof (clientparms_t));
		clientparms->clientsockaddr = calloc (1,
				sizeof(struct sockaddr_in));
		rc = comm_get_fd (ccc, &clientparms->clientsocket);
		clientparms->sockaddrsize = sizeof (struct sockaddr_in);
		rc = getpeername (clientparms->clientsocket,
				(struct sockaddr *)clientparms->clientsockaddr,
				&clientparms->sockaddrsize);
		clientparms->cc = ccc;
		// Validate the client
		// Hook the client connection
                rc = comm_group_add (Server.cg, ccc, clientparms, COMM_IO_READ);
                if (rc != 0) {
                        comm_close (&ccc, 0);
                        gxsmad_log_entry (
                                "comm_group_add(): failed rc=%d\n", rc);
                        return rc;
                }
                Server.conncount++;
                if ( (Server.conncount % 100) == 0) {
                        gxsmad_log_entry (
                                "Handled %lu connections so far.\n",
                                Server.conncount);
                }
                return rc;
        } else {
	        // It's a client socket, handle the request
		clientparms = (clientparms_t *)ccarg;
		if (clientparms == NULL) {
			gxsmad_log_entry ("%s: ccarg is NULL!\n", __PRETTY_FUNCTION__);
			return EINVAL;
		}

		// Do work here

		return rc;
	}
        return 0;
}



